(ffi:clines
"
void times_2(double *a, int len)
{
	int i;
	for (i=0; i<len; i++)
		a[i] = 2*a[i];
}
void print_int(int a)
{
	printf(\"value = %d\\n\", a);
}
int bar = 10;

typedef struct _S {
	int a;
	int b;
	int c;
} struct_S;
void print_struct(struct_S* s)
{
	printf(\"struct_S:\\n\");
	printf(\" a = %d\\n\", s->a);
	printf(\" b = %d\\n\", s->b);
	printf(\" c = %d\\n\", s->c);
}
struct_S foo = { 1, 2, 3 };
"
)

(ffi:def-function ("times_2" times2) ((a (* :double)) (len :int)) :returning :void)
(ffi:def-function ("print_int" print-int) ((a :int)) :returning :void)

(ffi:def-foreign-var ("bar" *bar*) :int "foo")

(ffi:def-struct struct-S
  (a :int)
  (b :int)
  (c :int))
(ffi:def-function ("print_struct" print-struct) ((s (* struct-S))) :returning :void)
(ffi:def-foreign-var ("foo" *foo*) struct-S "foo")

(print-int *bar*)
(terpri)

(ffi:with-foreign-object (s 'struct-S)
  (setf (ffi:get-slot-value s 'struct-S 'a) 5)
  (setf (ffi:get-slot-value s 'struct-S 'b) 10)
  (setf (ffi:get-slot-value s 'struct-S 'c) 15)
  (print-struct s)
  (terpri))

;;; Taken from arrays.lisp (examples of UFFI)
;;; ===> BEGIN <===

(defconstant +column-length+ 10)
(defconstant +row-length+ 10)

(ffi:def-foreign-type long-ptr (* :long))

(defun test-array-1d ()
  "Tests vector"
  (let ((a (ffi:allocate-foreign-object :long +column-length+)))
    (dotimes (i +column-length+)
      (setf (ffi:deref-array a '(:array :long) i) (* i i)))
    (dotimes (i +column-length+)
      (format t "~&~D => ~2D" i (ffi:deref-array a '(:array :long) i)))
    (ffi:free-foreign-object a))
  (values))

(defun test-array-2d ()
  "Tests 2d array"
  (let ((a (ffi:allocate-foreign-object 'long-ptr +row-length+)))
    (dotimes (r +row-length+)
      (declare (fixnum r))
      (setf (ffi:deref-array a '(:array (* :long)) r)
	    (ffi:allocate-foreign-object :long +column-length+))
      (let ((col (ffi:deref-array a '(:array (* :long)) r)))
	(dotimes (c +column-length+)
	  (declare (fixnum c))
	  (setf (ffi:deref-array col '(:array :long) c) (+ (* r +column-length+) c)))))

    (dotimes (r +row-length+)
      (declare (fixnum r))
      (format t "~&Row ~D: " r)
      (let ((col (ffi:deref-array a '(:array (* :long)) r)))
	(dotimes (c +column-length+)
	  (declare (fixnum c))
	  (let ((result (ffi:deref-array col '(:array :long) c)))
	    (format t "~2d " result)))))

    (ffi:free-foreign-object a))
  (values))

;;; ===> END <===

(test-array-1d)
(terpri) (terpri)
(test-array-2d)
(terpri) (terpri)

(defun print-vector (v n type)
  (format t "~&Vector: [ ")
  (dotimes (i n)
    (format t "~A " (ffi:deref-array v `(:array ,type) i)))
  (format t "]~%"))

(ffi:with-foreign-object (a `(:array :double ,+row-length+))
  (dotimes (i +row-length+)
    (setf (ffi:deref-array a '(:array :double) i) (* i i)))
  (print-vector a +row-length+ :double)
  (times2 a +row-length+)
  (print-vector a +row-length+ :double)
  (terpri))

(ffi:def-struct S1
  (a :int)
  (next :pointer-self))

(defun make-S1-atom (a &optional next)
  (unless next
    (setf next (ffi:make-null-pointer 'S1)))
  (let ((result (ffi:allocate-foreign-object 'S1)))
    (setf (ffi:get-slot-value result 'S1 'a) a)
    (setf (ffi:get-slot-value result 'S1 'next) next)
    result))

(defun print-S1 (obj)
  (format t "~&S1: { ")
  (do ((current obj (ffi:get-slot-value current 'S1 'next)))
      ((ffi:null-pointer-p current) nil)
    (format t "~A " (ffi:get-slot-value current 'S1 'a)))
  (format t "}~%"))

(let (current)
  (dotimes (i 10)
    (setq current (make-S1-atom i current)))
  (print-S1 current)
  (terpri))

(ffi:clines "#include <time.h>")

;;; Taken from gettime.lisp (examples from UFFI)
;;; ===> BEGIN <===

(ffi:def-foreign-type time-t :unsigned-long)

(ffi:def-struct tm
    (sec :int)
  (min :int)
  (hour :int)
  (mday :int)
  (mon :int)
  (year :int)
  (wday :int)
  (yday :int)
  (isdst :int))

(ffi:def-function ("time" c-time) 
    ((time (* time-t)))
  :returning time-t)

(ffi:def-function ("localtime" c-localtime)
    ((time (* time-t)))
  :returning (* tm))

(ffi:def-type time-t :unsigned-long)
(ffi:def-type tm-pointer (* tm))

(defun gettime ()
   "Returns the local time"
   (ffi:with-foreign-object (time 'time-t)
;;     (declare (type time-t time))
     (c-time time)
     (let ((tm-ptr (the tm-pointer (c-localtime time))))
       (declare (type tm-pointer tm-ptr))
       (let ((time-string (format nil "~2d/~2,'0d/~d ~2d:~2,'0d:~2,'0d" 
				  (1+ (ffi:get-slot-value tm-ptr 'tm 'mon))
				  (ffi:get-slot-value tm-ptr 'tm 'mday)
				  (+ 1900 (ffi:get-slot-value tm-ptr 'tm 'year))
				  (ffi:get-slot-value tm-ptr 'tm 'hour)
				  (ffi:get-slot-value tm-ptr 'tm 'min)
				  (ffi:get-slot-value tm-ptr 'tm 'sec)
				  )))
	 time-string))))

;;; ===> END <===

(format t "Time: ~S~%" (gettime))
(terpri)

#+msvc
(ffi:clines
"
#include <windows.h>
struct timezone {
  int tz_minuteswest;
  int tz_dsttime;
};
#define EPOCHFILETIME (0x019DB1DED53E8000)
int gettimeofday(struct timeval *tv, struct timezone *tz)
{
    FILETIME        ft;
    LARGE_INTEGER   li;
    LONGLONG         t;
    static int      tzflag;

    if (tv)
    {
        GetSystemTimeAsFileTime(&ft);
        li.LowPart  = ft.dwLowDateTime;
        li.HighPart = ft.dwHighDateTime;
        t  = li.QuadPart;       /* In 100-nanosecond intervals */
        t -= EPOCHFILETIME;     /* Offset to the Epoch time */
        t /= 10;                /* In microseconds */
        tv->tv_sec  = (long)(t / 1000000);
        tv->tv_usec = (long)(t % 1000000);
    }

    if (tz)
    {
        if (!tzflag)
        {
            _tzset();
            tzflag++;
        }
        tz->tz_minuteswest = _timezone / 60;
        tz->tz_dsttime = _daylight;
    }

    return 0;
}
")

;;; Taken from time.lisp (tests of UFFI)
;;; ===> BEGIN <===

(ffi:def-function "gmtime"
    ((time (* time-t)))
  :returning (* tm))

(ffi:def-function "asctime"
    ((time (* tm)))
  :returning :cstring)

(ffi:def-struct timeval 
    (secs :long)
  (usecs :long))

(ffi:def-struct timezone
    (minutes-west :int)
  (dsttime :int))

(ffi:def-function ("gettimeofday" c-gettimeofday) 
    ((tv (* timeval))
     (tz (* timezone)))
  :returning :int)

(defun get-utime ()
  (ffi:with-foreign-object (tv 'timeval)
    (let ((res (c-gettimeofday tv (ffi:make-null-pointer 'timezone))))
      (values
       (+ (* 1000000 (ffi:get-slot-value tv 'timeval 'secs))
	  (ffi:get-slot-value tv 'timeval 'usecs))
       res))))

(format t "get-utime test: ~A~%"
    (multiple-value-bind (t1 res1) (get-utime)
      (multiple-value-bind (t2 res2) (get-utime)
	(and (or (= t2 t1) (> t2 t1))
	     (> t1 1000000000)
	     (> t2 1000000000)
	     (zerop res1)
	     (zerop res2))))
)
(terpri)
	     
(defun posix-time-to-asctime (secs)
  "Converts number of seconds elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time (UTC)"
  (string-right-trim
   '(#\newline #\return)
   (ffi:convert-from-cstring
    (ffi:with-foreign-object (time 'time-t)
      (setf (ffi:deref-pointer time :unsigned-long) secs)
      (asctime (gmtime time))))))

(format t "posix-time-to-asctime:~%    result: ~A~%  expected: ~A~%"
    (posix-time-to-asctime #x7fffffff)
  "Tue Jan 19 03:14:07 2038")
(terpri)

;;; ===> END <===

;;; Taken from strtol.lisp (examples of UFFI)
;;; ===> BEGIN <===

(ffi:def-foreign-type char-ptr (* :unsigned-char))
  
;; This example does not use :cstring to pass the input string since
;; the routine needs to do pointer arithmetic to see how many characters
;; were parsed

(ffi:def-function ("strtol" c-strtol) 
    ((nptr char-ptr)
     (endptr (* char-ptr))
     (base :int))
  :returning :long)

(defun strtol (str &optional (base 10))
  "Returns a long int from a string. Returns number and condition flag.
Condition flag is T if all of string parses as a long, NIL if
their was no string at all, or an integer indicating position in string
of first non-valid character"
  (let* ((str-native (ffi:convert-to-foreign-string str))
	 (endptr (ffi:allocate-foreign-object 'char-ptr))
	 (value (c-strtol str-native endptr base))
	 (endptr-value (ffi:deref-pointer endptr 'char-ptr)))

    (unwind-protect
	 (if (ffi:null-pointer-p endptr-value)
	     (values value t)
	     (let ((next-char-value (ffi:deref-pointer endptr-value :unsigned-char))
		   (chars-parsed (- (ffi:pointer-address endptr-value)
				    (ffi:pointer-address str-native))))
	       (cond
		 ((zerop chars-parsed)
		  (values nil nil))
		 ((ffi:null-char-p next-char-value)
		  (values value t))
		 (t
		  (values value chars-parsed)))))
      (progn
	(ffi:free-foreign-object str-native)
	(ffi:free-foreign-object endptr)))))

;(progn
;  (flet ((print-results (str)
;	   (multiple-value-bind (result flag) (strtol str)
;	     (format t "~&(strtol ~S) => ~S,~S" str result flag))))
;    (print-results "55")
;    (print-results "55.3")
;    (print-results "a")))

;;; ===> END <===

(map nil #'(lambda (s)
             (multiple-value-bind (result flag) (strtol s)
	       (format t "(strtol ~S) => ~S,~S~%" s result flag)))
	 '("55" "55.3" "a"))
(terpri)

;;; Taken from atoifl.lisp (examples of UFFI)
;;; ===> BEGIN <===

(ffi:def-function ("atoi" c-atoi) 
    ((str :cstring))
  :returning :int)

(ffi:def-function ("atol" c-atol) 
    ((str :cstring))
  :returning :long)

(ffi:def-function ("atof" c-atof) 
    ((str :cstring))
  :returning :double)

(defun atoi (str)
  "Returns a int from a string."
  (ffi:with-cstring (str-cstring str)
    (c-atoi str-cstring)))

(defun atof (str)
  "Returns a double float from a string."
  (ffi:with-cstring (str-cstring str)
    (c-atof str-cstring)))

;;; ===> END <===

(let ((str "55"))
  (format t "(atoi ~S) => ~S~%" str (atoi str)))
(let ((str "2.23"))
  (format t "(atof ~S) => ~S~%" str (atof str)))
(terpri)

;;; Taken from casts.lisp (tests of UFFI)
;;; ===> BEGIN <===

(ffi:clines
"
void *
cast_test_int () {
  int *x = (int *) malloc(sizeof(int));
  *x = 23;
  return x;
}

void *
cast_test_float ()
{
  double *y = (double *) malloc(sizeof(double));
  *y = 3.21;
  return y;
}
")

(ffi:def-function ("cast_test_int" cast-test-int)
    ()
  :returning :pointer-void)

(ffi:def-function ("cast_test_float" cast-test-float)
    ()
  :returning :pointer-void)

(progn
  (ffi:with-cast-pointer (temp (cast-test-int) :int)
    (assert (= (ffi:deref-pointer temp :int) 23)))
  (let ((result (cast-test-int)))
    (ffi:with-cast-pointer (result2 result :int)
      (assert (= (ffi:deref-pointer result2 :int) 23)))
    (ffi:with-cast-pointer (temp result :int)
      (assert (= (ffi:deref-pointer temp :int) 23))))
  t)

(progn
  (ffi:with-cast-pointer (temp (cast-test-float) :double)
    (assert (= (ffi:deref-pointer temp :double) 3.21d0)))
  (let ((result (cast-test-float)))
    (ffi:with-cast-pointer (result2 result :double)
      (assert (= (ffi:deref-pointer result2 :double) 3.21d0)))
    (ffi:with-cast-pointer (temp result :double)
      (assert (= (ffi:deref-pointer temp :double) 3.21d0))))
  t)

;;; ===> END <===

(format t "Casts OK~%")
(terpri)

;;; Taken from getenv.lisp (tests of UFFI)
;;; (slightly modified because MSVC doesn't have (un)setenv)
;;; ===> BEGIN <===

(ffi:def-function ("getenv" c-getenv) 
    ((name :cstring))
  :returning :cstring)

(ffi:def-function ("putenv" c-putenv) 
    ((name :cstring))
  :returning :int)

(defun my-getenv (key)
  "Returns an environment variable, or NIL if it does not exist"
  (check-type key string)
  (ffi:with-cstring (key-native key)
    (ffi:convert-from-cstring (c-getenv key-native))))

(defun my-setenv (key name)
  "Returns an environment variable, or NIL if it does not exist"
  (check-type key string)
  (check-type name string)
  (ffi:with-cstrings ((env-native (concatenate 'string key "=" name)))
    (c-putenv env-native)))

(defun my-unsetenv (key)
  "Returns an environment variable, or NIL if it does not exist"
  (check-type key string)
  (ffi:with-cstrings ((key-native (concatenate 'string key "=")))
    (c-putenv key-native)))

(format t "__UFFI_FOO1__ = ~A~%"
  (progn
    (my-unsetenv "__UFFI_FOO1__")
    (my-getenv "__UFFI_FOO1__")))
(format t "__UFFI_FOO1__ = ~A~%"
  (progn
    (my-setenv "__UFFI_FOO1__" "UFFI-TEST")
    (my-getenv "__UFFI_FOO1__")))
(format t "__UFFI_FOO1__ = ~A~%"
  (progn
    (my-unsetenv "__UFFI_FOO1__")
    (my-getenv "__UFFI_FOO1__")))
(terpri)

;;; ===> END <===

;;; Taken from union.lisp (tests of UFFI)
;;; ===> BEGIN <===

(ffi:def-union tunion1 
  (char :char)
  (int :int)
  (uint :unsigned-int)
  (sf :float)
  (df :double))

(defvar *u* (ffi:allocate-foreign-object 'tunion1))
(setf (ffi:get-slot-value *u* 'tunion1 'uint)
      #-(or sparc sparc-v9 powerpc ppc)
      (+ (* 1 (char-code #\A))
	 (* 256 (char-code #\B))
	 (* 65536 (char-code #\C))
	 (* 16777216 128))
      #+(or sparc sparc-v9 powerpc ppc)
      (+ (* 16777216 (char-code #\A))
	 (* 65536 (char-code #\B))
	 (* 256 (char-code #\C))
	 (* 1 128)))

(format t "union.1: ~A =? ~A~%"
    (ffi:ensure-char-character 
     (ffi:get-slot-value *u* 'tunion1 'char))
  #\A)

(format t "union.2: ~A =? ~A~%"
    (ffi:ensure-char-integer 
     (ffi:get-slot-value *u* 'tunion1 'char))
  65)

#-(or sparc sparc-v9 mcl)
(format t "union.3: ~A =? ~A~%" (plusp (ffi:get-slot-value *u* 'tunion1 'uint)) t)

(terpri)

;;; ===> END <===

(ffi:load-foreign-library
  (ffi:find-foreign-library "testlib" "."))

(ffi:def-function ("f1" f1)
  ((a :int)
   (b :int))
  :returning :int)

(let ((a 2) (b 3))
  (format t "f1(~A,~A) = ~A~%" a b (f1 a b)))
(terpri)
